// 
// Copyright (c) 2002 David Eriksson <david@2good.nu>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// $Id: codegen.cpp,v 1.1.1.1 2002/06/20 18:00:44 twogood Exp $
#include "codegen.hpp"
#include "instruction.hpp"
#include "node.hpp"

/**
 * Instruction visitor for code generation
 */
class CodeGenerator : public InstructionVisitor
{
	public:
		CodeGenerator(CodeStyle style)
			: mStyle(style)
		{}

		virtual ~CodeGenerator()
		{
			msg("%s", mOut.str());
		}

		//
		// Implementation of InstructionVisitor interface follows
		//
		
		virtual void Visit(Assignment& instruction)
		{
			Prefix(instruction);
			if (!instruction.First()->IsType(Expression::DUMMY))
			{
				instruction.First()->GenerateCode(mOut);
				mOut << " = ";
			}
			instruction.Second()->GenerateCode(mOut);
			mOut << ';' << endl;
		}

		virtual void Visit(Case& instruction)
		{
			Prefix(instruction);
			mOut << "case " << instruction.Value() << ':' << endl;
		}

		virtual void Visit(ConditionalJump& instruction)
		{
			Prefix(instruction);
			mOut << "if (";
			instruction.First()->GenerateCode(mOut);
			mOut << ") goto ";
			instruction.Second()->GenerateCode(mOut);
			mOut << ';' << endl;
		}

		virtual void Visit(Jump& instruction)
		{
			Prefix(instruction);
			mOut << "goto ";
			instruction.Operand()->GenerateCode(mOut);
			mOut << ';' << endl;
		}

		virtual void Visit(Label& instruction)
		{
			Prefix(instruction, NO_INDENT);
			mOut << instruction.Name() << ':' << endl;
		}

		virtual void Visit(LowLevel& instruction)
		{
			Prefix(instruction);
			mOut << "/* Low-level instruction of type " 
				<< instruction.Insn().itype
				<< " */" << endl;
		}

		virtual void Visit(Push& instruction)
		{
			Prefix(instruction);
			mOut << "/* push ";
			instruction.Operand()->GenerateCode(mOut);
			mOut << " */" << endl;
		}

		virtual void Visit(Pop& instruction)
		{
			Prefix(instruction);
			mOut << "/* pop ";
			instruction.Operand()->GenerateCode(mOut);
			mOut << " */" << endl;
		}

		virtual void Visit(Return& instruction)
		{
			Prefix(instruction);
			mOut << "return ";
			instruction.Operand()->GenerateCode(mOut);
			mOut << ';' << endl;
		}

		virtual void Visit(Switch& instruction)
		{
			Prefix(instruction);
			mOut << "switch (";
			instruction.Operand()->GenerateCode(mOut);
			mOut << ')' << endl;
		}
		
		virtual void Visit(Throw& instruction)
		{
			Prefix(instruction);
			if (instruction.IsRethrow())
			{
				mOut << "throw;" << endl;
			}
			else
			{
				mOut << "throw ";
				instruction.Exception()->GenerateCode(mOut);
				mOut << "; // " << instruction.DataType() << endl;
			}
		}
		
		virtual void NodeEnd()
		{
			mOut << endl;
		}


	private:
		enum Indent
		{
			NO_INDENT, INDENT
		};

		void Prefix(Instruction& instruction, Indent indent = INDENT)
		{
			if (LISTING_STYLE == mStyle)
			{
				mOut.width(8);
				mOut.fill('0');
				mOut << hex << instruction.Address() << dec << ' ';
				mOut.width(0);
			}

			if (indent == INDENT)
				mOut << "  ";
		}

	private:

		CodeStyle mStyle;
		std::ostrstream mOut;
};

/**
 * Generate code for a list of instructions
 */
void GenerateCode(Instruction_list& instructions, CodeStyle style)
{
	CodeGenerator code_generator(style);
	Accept(instructions, code_generator);
}

/**
 * Generate code for list of nodes
 */
void GenerateCode(Node_list& nodes, CodeStyle style)
{
	CodeGenerator code_generator(style);
	Accept(nodes, code_generator);
}


