// 
// Copyright (c) 2002 David Eriksson <david@2good.nu>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// $Id: dataflow.cpp,v 1.1.1.1 2002/06/20 18:00:44 twogood Exp $
#include "dataflow.hpp"
#include "node.hpp"

bool DataFlowAnalysis::RemoveUnusedDefinition()/*{{{*/
{
	BoolArray& def = Instr()->Definitions();

	for (int reg = 0; reg < BoolArray::SIZE; reg++)
	{
		// Is register i is defined here?
		if (def.Get(reg))
		{
			// Are there no uses of this definition?
			if (Instr()->DefinitionHasNoUses(reg) &&
					!(Instr()->IsLastDefinition(reg) && 
						Node()->InLiveOut(reg)))
			{
				if (Instr()->RemoveDefinition(reg))
				{
					// We can remove the whole instruction!
//					msg("%p Wow! Removing instruction!\n", Instr()->Address());
					Erase(Iterator());
					return true;
				}
			}
		}		
	}
	return false;
}/*}}}*/

// ReplaceRegisterExpression {{{
class ReplaceRegisterExpressionHelper
{
	public:
		ReplaceRegisterExpressionHelper(
				unsigned short reg,
				Expression_ptr replacement)
			: mReplaceDone(false), mRegister(reg), mReplacement(replacement)
		{ }

		void Replace(Instruction_ptr instruction, int operand)
		{
			if (Replace(instruction->Operand(operand)))
			{
				instruction->Operand(operand, mReplacement);
				mReplaceDone = true;
			}
		}

		bool Replace(Expression_ptr parent)
		{
			if (parent->IsType(Expression::REGISTER) &&
					static_cast<Register*>(parent.get())->Index() == mRegister)
			{
				return true; /* tell caller to replace the parent! */
			}

			for (int i = 0; i < parent->SubExpressionCount(); i++)
			{
				if (Replace(parent->SubExpression(i)))
				{
					parent->SubExpression(i, mReplacement);
					mReplaceDone = true;
					return false;
				}
			}
			return false;
		}

		bool ReplaceDone() { return mReplaceDone; }

	private:
		bool mReplaceDone;
		unsigned short mRegister;
		Expression_ptr mReplacement;
};

bool ReplaceRegisterExpression(
		Instruction_ptr instruction,
		int operand,
		unsigned short reg,
		Expression_ptr replacement)
{
	ReplaceRegisterExpressionHelper helper(reg, replacement);
	helper.Replace(instruction, operand);
	return helper.ReplaceDone();
}/*}}}*/

class GetFunctionParametersFromStackHelper : public ExpressionVisitor/*{{{*/
{
	public:
		GetFunctionParametersFromStackHelper(DataFlowAnalysis* analysis)
			: mAnalysis(analysis)
		{}
	
		virtual void Visit(Call& expression)
		{
			mAnalysis->CollectParameters(&expression);
		}

		virtual void Visit(BinaryExpression&)  {}
		virtual void Visit(Dummy&)             {}
		virtual void Visit(Global&)            {}
		virtual void Visit(NumericLiteral&)    {}
		virtual void Visit(Register&)          {}
		virtual void Visit(StackVariable&)     {}
		virtual void Visit(StringLiteral&)     {}
		virtual void Visit(TernaryExpression&) {}
		virtual void Visit(UnaryExpression&)   {}
		
	private:
		DataFlowAnalysis* mAnalysis;
}; /*}}}*/

void DataFlowAnalysis::CollectParameters(Call* call)/*{{{*/
{
	if (call->IsFinishedAddingParameters())
		return; // already collected parameters for this call 
	
	int parameters_left = call->ParameterCount();

	if (Call::UNKNOWN_PARAMETER_COUNT == parameters_left)
	{
		parameters_left = Stack().size();
		msg("%p I guess this function call takes %i parameters.\n", 
				Instr()->Address(), parameters_left);
		call->ParameterCountFromStack(parameters_left);
	}

	while (!Stack().empty() && parameters_left > 0)
	{
		Push* push = static_cast<Push*>(Stack().top()->get());
		call->AddParameter( push->Operand() );
		Erase(Stack().top());
		Stack().pop();
		parameters_left--;
	}

	call->SetFinishedAddingParameters();

	if (parameters_left != 0)
	{
		msg("%p Unexpected number of parameters left: %i. Wanted %i parameters.\n", 
				Instr()->Address(), parameters_left, call->ParameterCount());
	}

#if 0
	call->SetDataTypes();
#endif
} /*}}}*/

void DataFlowAnalysis::GetFunctionParametersFromStack()/*{{{*/
{
	GetFunctionParametersFromStackHelper helper(this);
	
	for (int i = 0; i < Instr()->OperandCount(); i++)
	{
		Instr()->Operand(i)->AcceptDepthFirst(helper);
	}
}/*}}}*/

void DataFlowAnalysis::TryConvertPushPopToAssignment()/*{{{*/
{
	Instruction_list::iterator pop  = Iterator();
	Instruction_list::iterator push = Stack().top();

	if (Stack().size() == 0)
	{
		msg("%p [TryConvertPushPopToAssignment] Empty stack\n", Instr()->Address());
		return;
	}

	if (Instructions().end() == pop)
	{
		msg("%p [TryConvertPushPopToAssignment] Bad pop\n", Instr()->Address());
		return;
	}

	if (Instructions().end() == push)
	{
		msg("%p [TryConvertPushPopToAssignment] Bad push\n", Instr()->Address());
		return;
	}
	
	//
	//  If there are no defintions of popped register between push and pop,
	//  we can make an assignment at the push
	//
	Expression_ptr popped = (**pop).Operand(0);

	// This should always be a register, but who knows?
	if (popped->IsType(Expression::REGISTER))
	{
		Expression_ptr pushed = (**push).Operand(0);

		unsigned short reg = static_cast<Register*>(pushed.get())->Index();
		Instruction_list::iterator i = push;

		for (i++; i != pop; i++)
		{
			if ( (**i).Definitions().Get(reg))
				break;
		}

		if (i == pop)
		{
			msg("%p Replacing push/pop with assignment!\n", 
					(**push).Address());
			Instructions().insert(
					push,
					Instruction_ptr(new Assignment(
							(**push).Address(),
							popped,	
							pushed
							))
					);
			Erase(push);
			Erase(pop);
		}
	}
}/*}}}*/

Instruction_list::iterator DataFlowAnalysis::FindInstructionAtAddress(/*{{{*/
		Instruction_list::iterator item, ea_t address)
{
	for(item++; item != Instructions().end(); item++)
	{
		if ((**item).Address() == address)
			break;
	}
	return item;
}/*}}}*/

Analysis::AnalysisResult DataFlowAnalysis::ReplaceUseWithDefinition(/*{{{*/
		Assignment* assignment)
{
	RegisterToAddress_map& du_chain = assignment->DuChain();

	if (du_chain.size() != 1)
		return CONTINUE;

	unsigned short reg = du_chain.begin()->first;
//	msg("%p du_chain.count(%i) = %i\n", assignment->Address(), reg, du_chain.count(reg));

	// Don't do this for the last defintion if it is in LiveOut
	if (assignment->IsLastDefinition(reg) && Node()->InLiveOut(reg))
		return CONTINUE;

#if 0
	pair<RegisterToAddress_map::const_iterator, RegisterToAddress_map::const_iterator> range = 
		du_chain.equal_range(reg);

	for (RegisterToAddress_map::const_iterator item = range.first;
			item != range.second;
			item++)
	{
		msg("item->second = %p\n", item->second);
	}
#endif

	// Find end of DU-chain
	Instruction_list::iterator target_item = 
		FindInstructionAtAddress(Iterator(), du_chain.begin()->second);

	// Verify we found it
	if (Instructions().end() == target_item)
		return CONTINUE;

	// XXX: not for calls?
	if (assignment->IsCall())
	{
		// XXX: See if adjacent instructions
		Instruction_list::iterator tmp = Iterator();
		tmp++;
		if (tmp != target_item)
			return CONTINUE;
	}

	Instruction_ptr target = *target_item;
	bool replace_done = false;

	// assuming the register is only used once
	for (int i = 0; i < target->OperandCount(); i++)
	{
		if (target->OperandType(i) == Instruction::USE)
		{
			if (ReplaceRegisterExpression(target, i, reg, assignment->Second()))
			{
				replace_done = true;
				break;
			}
		}
	}

	if (replace_done)
	{
		target->Uses().Clear(reg);
		target->Uses() |= assignment->Uses();
		target->LastDefinitions() |= assignment->LastDefinitions();
		Erase(Iterator());
	}

	return CONTINUE;
}/*}}}*/

Analysis::AnalysisResult DataFlowAnalysis::ReplaceUseWithDefinition2(/*{{{*/
		Assignment* assignment)
{
	// Must be a register
	if (!assignment->First()->IsType(Expression::REGISTER))
		return CONTINUE;
	
	unsigned short reg = Register::Index(assignment->First());

	// Must be last definition and be in LiveOut
	if (!assignment->IsLastDefinition(reg) ||
			!Node()->InLiveOut(reg))
		return CONTINUE;

	// Must be a primitive type
	switch (assignment->Second()->Type())
	{
		case Expression::GLOBAL:
		case Expression::NUMERIC_LITERAL:
		case Expression::STACK_VARIABLE:
		case Expression::STRING_LITERAL:
			break;

		default:
			return CONTINUE;
	}

/*	msg("%p In ReplaceUseWithDefinition2 and past the tests\n", 
			assignment->Address());*/

	// for each successor
	for (int i = 0; i < Node()->SuccessorCount(); i++)
	{
		Node_ptr successor = Node()->Successor(i);
		if (!successor->LiveIn().Get(reg))
			continue;
//		msg("  sucessor->Address() = %p\n", successor->Address());
		if (successor->InLiveOut(reg))
		{
//			msg("Can't handle substitution on more than the direct successors yet :-(\n");
			return CONTINUE;
		}
	
		for (Node_list::iterator n = mNodeList.begin();
				n != mNodeList.end();
				n++)
		{
			Node_ptr node = *n;

			if (Node() != node)
			{
				for (int i = 0; i < node->SuccessorCount(); i++)
				{
					if (node->Successor(i) == successor)
					{
//						msg("    predecessor found @ %p\n", node->Address());
						if (node->InLiveOut(reg))
						{
//							msg("      predecessor had register in LiveOut :-(\n");
							return CONTINUE;
						}
					}
				}
			}
		}
	}

	// for each successor again
	for (int i = 0; i < Node()->SuccessorCount(); i++)
	{
		Node_ptr successor = Node()->Successor(i);

		if (!successor->LiveIn().Get(reg))
			continue;
		
		// for each instruction in successor
		for (Instruction_list::iterator target_item = successor->Instructions().begin();
				target_item != successor->Instructions().end();
				target_item++)
		{
			Instruction_ptr target = *target_item;

#if 1
			bool replace_done = false;

			// assuming the register is only used once
			for (int i = 0; i < target->OperandCount(); i++)
			{
				if (target->OperandType(i) == Instruction::USE)
				{
					if (ReplaceRegisterExpression(target, i, reg, assignment->Second()))
					{
						replace_done = true;
						break;
					}
				}
			}

			if (replace_done)
			{
//				msg("%p Replaced expression in this instruction\n", target->Address());

				target->Uses().Clear(reg);
				target->Uses() |= assignment->Uses();
				//target->LastDefinitions() |= assignment->LastDefinitions();
			}

			// Stop if the register was redefined by this instruction
			if (target->Definitions().Get(reg))
				break;
		}
			
		successor->LiveIn().Clear(reg);
	}
	
	Node()->LiveOut().Clear(reg);
//	Erase(Iterator());
#endif

	return CONTINUE;			
}/*}}}*/

void DataFlowAnalysis::TryBorlandThrow(Assignment* assignment)/*{{{*/
{
	if (assignment->Second()->IsType(Expression::CALL))
	{
		Call* call = static_cast<Call*>(assignment->Second().get());

		// 1 function + 9 parameters
		if (call -> SubExpressionCount() == 10 &&
				call->SubExpression(0)->IsType(Expression::GLOBAL) &&
				call->SubExpression(1)->IsType(Expression::GLOBAL) &&
				call->SubExpression(2)->IsType(Expression::UNARY_EXPRESSION))
		{
			Global* global = static_cast<Global*>(call->SubExpression(0).get());

			if (global->Name() == "@_ThrowExceptionLDTC$qpvt1t1t1uiuiuipuct1")
			{
				// This seems to be a Borland C++ throw instruction
				
				msg("%p Borland C++ throw instruction\n", Instr()->Address());
				
				//
				// Parameters:
				//
				// 1: RTTI data about the type we throw
				// 2: Address of the exception value/pointer we throw
				// 3: 0
				// 4: 0 or pointer to constructor
				// 5: 0 or 1
				// 6: 0
				// 7: 0
				// 8: 0
				// 9: 0 or some value
				//

				Global* rtti = 
					static_cast<Global*>(call->SubExpression(1).get());
				ea_t offset = rtti->Address();

				UnaryExpression* exception_address = 
					static_cast<UnaryExpression*>(call->SubExpression(2).get());
				
				if (0 == strcmp(exception_address->Operation(), "&") &&
						BADADDR != offset)
				{
					flags_t flags = getFlags(offset);
					flags_t type_flags;
					ea_t type_offset = offset + get_word(offset + 6);

					if (isStruct(flags))
					{
						type_flags = get_item_flag(BADADDR, 0, type_offset, 0);
					}
					else
					{
						if (isData(flags))
						{
							do_unknown(type_offset, true);
						}
						
						// TODO: convert to struct, not just string
						make_ascii_string(type_offset, 0, ASCSTR_C);
						type_flags = getFlags(type_offset);
					}

					if (isASCII(type_flags))
					{
						ulong string_type = get_str_type(type_offset);
						string data_type = 
							StringLiteral::GetString(type_offset, string_type);

						Insert(
								new Throw(
									Instr()->Address(), 
									exception_address->Operand(), 
									data_type));
						
						Erase(Iterator());
					}
				}
			}

		}
		else if (call->SubExpressionCount() == 3 &&
				call->SubExpression(0)->IsType(Expression::GLOBAL))
		{
			Global* global = static_cast<Global*>(call->SubExpression(0).get());

			if (global->Name() == "@_ReThrowException$quipuc")
			{
				msg("%p Borland C++ rethrow instruction\n", Instr()->Address());
				Insert( new Throw( Instr()->Address() ) );
				Erase(Iterator());
			}
		}

	}
}/*}}}*/

void DataFlowAnalysis::OnAssignment(Assignment* assignment)/*{{{*/
{
#if 0
	if (Expression::Equal(assignment->First(), assignment->Second()))
	{
		msg("%p Assign to self\n", Instr()->Address());
	}
#endif

	//ReplaceUseWithDefinition2(assignment);
	ReplaceUseWithDefinition(assignment);
	//TryIncDec(assignment);

	// Propagate data type
	assignment->First()->DataType() = assignment->Second()->DataType();

	TryBorlandThrow(assignment);
}/*}}}*/


