// 
// Copyright (c) 2002 David Eriksson <david@2good.nu>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// $Id: expression.cpp,v 1.1.1.1 2002/06/20 18:00:45 twogood Exp $

//
// C++ headers
//
#include <strstream>

//
// Boost headers
//
#include <boost/shared_array.hpp>

//
// IDA headers
//
#include <ida.hpp>
#include <idp.hpp>
#include <typeinf.hpp>
#include <funcs.hpp>
#include <struct.hpp>
#include <frame.hpp>

//
// Local headers
//
#include "instruction.hpp"
#include "expression.hpp"
#include "x86.hpp"
	
using namespace std;

const char* Register::mName_x86_16bit[] = 
{
	"ax","cx","dx","bx","sp","bp","si","di",
	"al","cl","dl","bl","ah","ch","dh","bh"
};

const char* Register::Name(int index)/*{{{*/
{
	if (index >= 0 && (index < (sizeof(mName_x86_16bit)/sizeof(char*))))
		return mName_x86_16bit[index];
	else
	{
		strstream os;
		os << "REG_" << hex << index << ends;
		return strdup(os.str()); // XXX: memory leak
	}
}/*}}}*/

Expression_ptr RegTimesValue(unsigned short reg, unsigned long value)/*{{{*/
{
	switch (value)
	{
		case 0:
			return NumericLiteral::Create(0);

		case 1:
			return Register::Create(reg);
			
		default:
			return Expression_ptr( new BinaryExpression(
						Register::Create(reg),
						"*",
						NumericLiteral::Create(value)
						));
	}
}/*}}}*/

Expression_ptr RegPlusReg(unsigned short reg1,unsigned short reg2)/*{{{*/
{
	return Expression_ptr( new BinaryExpression(
				Register::Create(reg1),
				"+",
				Register::Create(reg2)
				));
}/*}}}*/

Expression_ptr RegPlusRegTimesValue(unsigned short reg1,unsigned short reg2, unsigned long value)/*{{{*/
{
	if (0 == value)
		return Register::Create(reg1);
	else
		return Expression_ptr( new BinaryExpression(
					Register::Create(reg1),
					"+",
					RegTimesValue(reg2, value)
					));
}/*}}}*/

/* 
	 mov     [eax+edx*2+14h], di
	 Operands[0]={type=4, dtyp=1, reg/phrase=4, value=0, addr=20, sib=00000050}
	 o_displ               ???       eax?     14h       eax+edx*2
	 */

Expression_ptr GetSibExpression(insn_t& insn, int operand)/*{{{*/
{
	Expression_ptr result;
	
	unsigned char sib = insn.Operands[operand].sib & 0xff;
	
#if 0
	switch (sib & 0xff)
	{
		case 0x02:
			result = RegPlusReg(REG_DX, REG_AX);
			break;

		case 0x0a:
			result = RegPlusReg(REG_DX, REG_CX);
			break;

		case 0x0f:
			result = RegPlusReg(REG_DI, REG_CX);
			break;

		case 0x10:
			result = RegPlusReg(REG_AX, REG_DX);
			break;

		case 0x16:
			result = RegPlusReg(REG_SI, REG_DX);
			break;

		case 0x19:
			result = RegPlusReg(REG_CX, REG_BX);
			break;

		case 0x2b:
			result = RegPlusReg(REG_BX, REG_BP);
			break;

		case 0x31:
			result = RegPlusReg(REG_CX, REG_SI);
			break;

		case 0x32:
			result = RegPlusReg(REG_DX, REG_SI);
			break;

		case 0x33:
			result = RegPlusReg(REG_BX, REG_SI);
			break;

		case 0x41: // [ecx+eax*2]
			result = RegPlusRegTimesValue(REG_CX, REG_AX, 2);
			break;
		
		case 0x50: // [eax+edx*2]
			result = RegPlusRegTimesValue(REG_AX, REG_DX, 2);
			break;

		case 0x51: // [ecx+edx*2]
			result = RegPlusRegTimesValue(REG_CX, REG_DX, 2);
			break;

		case 0x71: // [ecx+esi*2]
			result = RegPlusRegTimesValue(REG_CX, REG_SI, 2);
			break;
				
		case 0x85: // [eax*4]
			result = RegTimesValue(REG_AX, 4);
			break;

		case 0xb6: // [esi+esi*4] -> esi*5
			result = RegTimesValue(REG_SI, 5);
			break;

		default:
			msg("%p Unknown SIB: %p\n", insn.ea, op.sib);
			result.reset( new Dummy() );
			break;
	}
#endif

  unsigned short scale = 1 << (sib >> 6);
  unsigned short index_reg = (sib >> 3) & 0x07;
  unsigned short base_reg = sib & 0x07;

	if (base_reg == 5)
	{
		result = RegTimesValue(index_reg, scale);

		msg("%p SIB: base=%i, %s*%i\n", insn.ea, 
				base_reg, Register::Name(index_reg), scale);
	}
	else
	{
		result = RegPlusRegTimesValue(base_reg, index_reg, scale);
	}

	return result;
}/*}}}*/

Expression_ptr Expression::FromOperand(insn_t& insn, int operand,/*{{{*/
				TypeInformation* type)
{
	Expression_ptr result;

	flags_t flags = ::getFlags(insn.ea);
	if (::isStkvar(flags, operand))
	{
		result = StackVariable::CreateFrom(insn, operand);
		if (NN_lea == insn.itype)
		{
			result.reset( new UnaryExpression("&", result) );
		}
	}
	else if (::isEnum(flags, operand))
	{
		msg("%p TODO: handle enum!!\n", insn.ea);
	}
	
	if (!result.get())
	{
		op_t op = insn.Operands[operand];
		refinfo_t refinfo;

#if 0
		if (op.hasSIB)
		{
			msg("%p Operand %i of type %i has SIB=%p\n", 
					insn.ea, operand, op.type, op.sib);
		}
#endif

		switch (op.type)
		{
			case o_reg:
				result.reset( new Register(op.reg) );
				break;

			case o_imm:
				if (::get_refinfo(insn.ea, operand, &refinfo))
				{
					ea_t address = refinfo.base + op.value;
					flags_t flags = ::getFlags(address);

					if (isASCII(flags))
					{
						result = StringLiteral::CreateFrom(address);
						break;
					}
					else
					{
						// XXX: maybe use & operator for result?
						result = Global::CreateFrom(address, insn.ea);
						if (result.get())
							break;
					}
				}
				result.reset( new NumericLiteral(op.value) );
				break;

			case o_near:
				if (!(insn.auxpref & aux_ad_is_32))
				{
					// Copy current segment value for 16-bit adresses
					op.addr |= insn.ea & 0xffff0000;
				}
				// fall though
				
			case o_mem:
			case o_far:
				result = Global::CreateFrom(op.addr, insn.ea);
				if (result.get())
				{
					if (NN_lea == insn.itype)
					{
						result.reset(
								new UnaryExpression(
								"&",
								result));
					}
				}

				if (op.hasSIB)
				{
					if (!result.get())
					{
						result.reset(new NumericLiteral(op.addr));
					}

					result.reset(
							new BinaryExpression(
								result,
								"+",
								GetSibExpression(insn, operand)
								));
					break;
				}

				if (!result.get())
				{
					if (0x14 == op.segrg && 0 == op.addr)
					{
						 // Operand is "large fs:0"
						 msg("%p This instruction uses the FS segment register and should probably be handled special\n", insn.ea);
					}
					else
					{
						msg("%p [Expression::FromOperand] UNEXPECTED MEMORY ADDRESS", insn.ea);
						msg(", phrase=%i, addr=%08x, value=%i, specval=%x\n", 
								op.phrase, op.addr, op.value, op.specval);
					}
					result.reset( new Dummy() );
				}
				break;

			case o_phrase:
				// fall through

			case o_displ:
				if (op.hasSIB)
				{
					result = GetSibExpression(insn, operand);
				}
				else
				{
					unsigned short reg = (unsigned short)-1;

				
					switch (op.phrase)
					{
						case 0: reg = REG_AX; break;
						case 1: reg = REG_CX; break;
						case 2: reg = REG_DX; break;
						case 3: reg = REG_BX; break;
						case 6: reg = REG_SI; break;
						case 7: reg = REG_BX; break;

						default:
							msg("%p Warning! CHECK OUT OPERAND %i of type %i. op.phrase=%i\n", 
									insn.ea, operand, op.type, op.phrase);
							break;
					}
							
					result.reset(new Register(reg));
				}

				if (0 != op.addr)
				{
					result.reset( new BinaryExpression(
								result,
								"+",
								Expression_ptr( new NumericLiteral(op.addr) ))
							);
				}

				if (NN_lea != insn.itype)
				{
					result.reset( new UnaryExpression("*", result) );
				}
				break;

			default:
				msg("%p Unknown operand type %i\n", insn.ea, op.type);
				result.reset( new Dummy() );
				break;
		}
	}

	if (result.get() && NULL != type)
	{
		result->DataType() = *type;
	}
	
	return result;
}/*}}}*/

Call::Call(Expression_ptr function)/*{{{*/
	: Expression(CALL), mFunctionAddress(BADADDR),
		mParameterCount(UNKNOWN_PARAMETER_COUNT), 
		mCallingConvention(CM_CC_UNKNOWN),
		mFinishedAddingParameters(false)
{
	memset(mDataTypes,      0, sizeof(mDataTypes));
	memset(mParameterNames, 0, sizeof(mParameterNames));
//	memset(mReturnType,     0, sizeof(mReturnType));

	mSubExpressions.push_back(function);
	
	if (function->IsType(GLOBAL))
	{
		// Get address from name of function
		mFunctionAddress = static_cast<Global*>(function.get())->Address();
	}

	LoadFunctionInformation();
	LoadTypeInformation();
}/*}}}*/

Call::~Call()/*{{{*/
{
	free_funcarg_arrays(mDataTypes, mParameterNames, MAX_PARAMETERS);
}/*}}}*/

void Call::LoadFunctionInformation()/*{{{*/
{
	if (BADADDR == mFunctionAddress)
		return;
	
	func_t* func = get_func(mFunctionAddress);
	if (func)
	{
		if (func->argsize > 0)
		{
//			msg("Function at %p purges %i bytes\n", mFunctionAddress, func->argsize);

			// XXX: this is for 32-bit code
			ParameterCount(func->argsize >> 2);
		}
	}
	else
	{
		ulong purge = get_ind_purged(mFunctionAddress);
		if (purge != (ulong)-1)
		{
//			msg("Function at %p purges %i bytes\n", mFunctionAddress, purge);
		
			// XXX: this is for 32-bit code
			ParameterCount(purge >> 2);
		}
	}
}/*}}}*/

#if 0
int mymsg(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	int result = vmsg(format, ap);
	va_end(ap);
	return result;
}
#endif

void Call::LoadTypeInformation()/*{{{*/
{
	if (BADADDR == mFunctionAddress)
		return;

	type_t type[MAXSTR];
	p_list names[MAXSTR];
	
	if (!get_ti(mFunctionAddress, type, names))
		return;

	// CM (calling convention & model)
	cm_t cm = type[1];
	CallingConvention(cm);

	ulong plocations[MAX_PARAMETERS];
	memset(plocations, 0, sizeof(plocations));

	ParameterCount(
			build_funcarg_arrays(type, names, plocations, 
			(type_t**)mDataTypes, (char**)mParameterNames, MAX_PARAMETERS, false));

#if 0
	char buffer[MAXSTR]; 
	for (int i = 0; i < mParameterCount; i++)
	{
		buffer[0] = '\0';
		print_type_to_one_line(buffer, sizeof(buffer), idati, mDataTypes[i]);
		msg("Parameter %i type: %s %s\n", i, buffer, mParameterNames[i]);
	}
#endif

	type_t return_type[MAXSTR];
	if (::extract_func_ret_type(type, return_type, sizeof(return_type)))
	{
		DataType().Set(return_type);
#if 0
		buffer[0] = '\0';
		print_type_to_one_line(buffer, sizeof(buffer), idati, mReturnType);
		msg("Return type: %s\n", buffer);
#endif
	}
}/*}}}*/

ea_t DataSeg()
{
	for(int i = 0; i < get_segm_qty(); i++)
	{
		segment_t* s = getnseg(i);
		if (SEG_DATA == s->type)
			return get_segm_base(s);
	}
	return 0;
}

#if 0
void Call::SetDataTypes()/*{{{*/
{
	ea_t base = DataSeg();
	
	for (int i = 0; i < mParameterCount; i++)
	{
		if (mDataTypes[i])
		{
			Expression_ptr e = SubExpression(i+1);

#if 0
			char buffer[MAXSTR]; 
			buffer[0] = '\0';
			print_type_to_one_line(buffer, sizeof(buffer), idati, mDataTypes[i]);
			msg("Parameter %i type: %s\n", i, buffer);
#endif
			
			e->DataType().Set(mDataTypes[i]);
			
			if (e->IsType(Expression::NUMERIC_LITERAL) &&
					e->DataType().IsCharPointer())
			{
				// This is an immediate that is a string pointer... aha!
				Expression_ptr string = StringLiteral::CreateFrom(
						base + static_cast<NumericLiteral*>(e.get())->Value() 
						);

				if (string.get())
				{
					string->DataType().Set(mDataTypes[i]);
					SubExpression(i+1, string);
				}
			}
		}
	}
}/*}}}*/
#endif

Expression_ptr StringLiteral::CreateFrom(ea_t address)/*{{{*/
{
	Expression_ptr result;
	
	ulong type = get_str_type(address);

	string value = GetString(address, type);
	if (!value.empty())
		result.reset(new StringLiteral(value, type));

	return result;
}/*}}}*/

string StringLiteral::GetString(ea_t address, ulong type)/*{{{*/
{
	size_t len = get_max_ascii_length(address, type, false);
	boost::shared_array<char> str(new char[len+1]);
	get_ascii_contents(address, len, type, str.get(), len+1);
	return str.get();
}/*}}}*/

string StringLiteral::EscapeAsciiString(const std::string& str)/*{{{*/
{
	return EscapeAsciiString(str.c_str());
}/*}}}*/

string StringLiteral::EscapeAsciiString(const char* str)/*{{{*/
{
	int escape_count = 0;
	const char* p = str;
		
	while (*p)
	{
		if ('\n' == *p)
			escape_count++;

		p++;
	}

	if (escape_count)
	{
		boost::shared_array<char> new_str(new char[strlen(str) + escape_count + 1]);
		char* q = new_str.get();
		p = str;

		while (*p)
		{
			if ('\n' == *p)
			{
				*q++ = '\\';
				*q++ = 'n';
				p++;
			}
			else
			{
				*q++ = *p++;
			}
		}

		*q = 0;
//		msg("New string: %s\n", new_str);
		return new_str.get();
	}
	else
	{
		return str;
	}
}/*}}}*/


/**
 * May return a NULL pointer
 */
Expression_ptr Global::CreateFrom(ea_t ea, ea_t from)/*{{{*/
{
	Expression_ptr result;
	
	string name = GetName(ea, from);
	if (!name.empty())
		result.reset(new Global(name, /*index*/0, ea));

	return result;
}/*}}}*/

string Global::GetName(ea_t ea, ea_t from)/*{{{*/
{
	char name[MAXSTR];
	if (get_name(from, ea, name, sizeof(name)))
	{
		return string(name);
	}
	else
	{
		return string("");
	}
}/*}}}*/

Expression_ptr StackVariable::CreateFrom(insn_t& insn, int operand)/*{{{*/
{
	Expression_ptr result;
	
	int index;
	string name = GetName(insn.ea, operand, &index);

	if (name.empty())
	{
		// Try to add a stack variable and try again!
//		msg("%p Warning: trying to create stack variable\n", insn.ea);
		add_stkvar(insn.Operands[operand], insn.Operands[operand].addr);
		op_stkvar(insn.ea, operand);
		name = GetName(insn.ea, operand, &index);
	}
	
	if (!name.empty())
		result.reset(new StackVariable(name, index));

	return result;
}/*}}}*/

string StackVariable::GetName(ea_t ea, int operand, int* pIndex)/*{{{*/
{
	ostrstream buffer;
	bool first = true;

	if (pIndex)
		*pIndex = 0;
	
	func_t* func = get_func(ea);
	if (func)
	{
		ulong offset = calc_stkvar_struc_offset(func, ea, operand);

		struc_t* struc = get_frame(func);
		member_t* member;  /* must live here, not inside while */

		while (struc)
		{
			member = get_member(struc, offset);
			if (member)
			{
				if (first)
					first = false;
				else
					buffer << '.';

				char* tmp = get_member_name(member->id);
				if (tmp)
					buffer << get_member_name(member->id);
				else
					buffer << "NO_NAME";

				struc = get_sptr(member);
			}
			else
			{
				struc = NULL;
			}
		}

		if (member && (offset - member->soff) != 0)
		{
			//msg("%p Stack variable: offset=%i, member->soff=%i\n", ea, offset, member->soff);
			if (pIndex)
				*pIndex = offset - member->soff;
		}

	}

	buffer << ends;

	return buffer.str();
}/*}}}*/

#if 1
bool Expression::Equal(Expression_ptr a, Expression_ptr b)
{
//	msg("Comparing two expressions\n");
	
	if (a->Type() != b->Type() ||
			a->SubExpressionCount() != b->SubExpressionCount())
		return false;

	for (int i = 0; i < a->SubExpressionCount(); i++)
	{
		if (!Equal(a->SubExpression(i), b->SubExpression(i)))
			return false;
	}

	switch (a->Type())
	{
		case UNARY_EXPRESSION:
		case BINARY_EXPRESSION:
		case TERNARY_EXPRESSION:
			// TODO: check operators
			return false;

		case REGISTER:
			return 
				static_cast<Register*>(a.get())->Index() ==
				static_cast<Register*>(b.get())->Index();

		case NUMERIC_LITERAL:
			return 
				static_cast<NumericLiteral*>(a.get())->Value() ==
				static_cast<NumericLiteral*>(b.get())->Value();

			// TODO: handle all types
	}

	return false;
}
#endif

