// 
// Copyright (c) 2002 David Eriksson <david@2good.nu>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// $Id: expression.hpp,v 1.1.1.1 2002/06/20 18:00:45 twogood Exp $
#ifndef _EXPRESSION_HPP
#define _EXPRESSION_HPP

#include "desquirr.hpp"

class Global;

class TypeInformation/*{{{*/
{
	public:
		TypeInformation()
		{
			mDataType = (type_t*)qalloc(1);
			mDataType[0] = BT_UNK;
		}

		TypeInformation& operator=(TypeInformation& other)
		{
			Set(other.mDataType);	
			return *this;
		}

		bool operator==(TypeInformation& other)
		{
			return ::equal_types(::idati, mDataType, other.mDataType);
		}
		
		TypeInformation(type_t* dataType)
		{
			Set(dataType);
		}

		~TypeInformation()
		{
			Clear();
		}

		void Set(type_t* dataType) 
		{
			Clear();
			mDataType = ::typdup(dataType);
		}

		void Clear()
		{
			::qfree(mDataType);
			mDataType = NULL;
		}

		void Write(ostream& os)/*{{{*/
		{
			char buffer[MAXSTR]; 
			buffer[0] = '\0';
			::print_type_to_one_line(buffer, sizeof(buffer), ::idati, mDataType);

			os << buffer;
		}/*}}}*/

		bool IsVoid(int i = 0)
		{ return ::is_type_void  (mDataType[i]); }
		
		bool IsPointer(int i = 0)  
		{ return ::is_type_ptr   (mDataType[i]); }
		
		type_t BaseType(int i = 0) 
		{ return ::get_base_type (mDataType[i]); }
		
		type_t Flags(int i = 0)    
		{ return ::get_type_flags(mDataType[i]); }

		void Flags(type_t flags, int i = 0)    
		{ 
			mDataType[i] &= ~TYPE_FLAGS_MASK;
			mDataType[i] |= (flags & TYPE_FLAGS_MASK);
		}

		int Size() { return ::typlen(mDataType); }

		void Signness(type_t signness, int i = 0)/*{{{*/
		{
			switch (BaseType(i))
			{
				case BT_INT8:
				case BT_INT16:
				case BT_INT32:
				case BT_INT64:
				case BT_INT128:
				case BT_INT:
					Flags(signness, i);
					break;
			}
		}/*}}}*/

		type_t Signness(int i = 0)/*{{{*/
		{
			switch (BaseType(i))
			{
				case BT_INT8:
				case BT_INT16:
				case BT_INT32:
				case BT_INT64:
				case BT_INT128:
				case BT_INT:
					return Flags(i);

				default:
					return BTMT_UNKSIGN;
			}
		}/*}}}*/

		int PointerHeaderSize(int i = 0)/*{{{*/
		{
			if (IsPointer())
				return ::skip_ptr_type_header(mDataType + i) - (mDataType + i);
			else
				return 0;
		}/*}}}*/

		bool IsCharPointer()/*{{{*/
		{
			if (IsPointer())
			{
				int i = PointerHeaderSize();	
				return BaseType(i) == BT_INT8 && Flags(i) == BTMT_CHAR;
			}
			return false;
		}/*}}}*/

		void MakeCharPointer()
		{
			// TODO
		}

		void MakeBasicType(type_t type)/*{{{*/
		{
			Clear();
			mDataType = (type_t*)qalloc(2);
			mDataType[0] = type;
			mDataType[1] = 0;
		}/*}}}*/
		
		void MakeInt(type_t signness = BTMT_UNKSIGN)/*{{{*/
		{
			MakeBasicType(BT_INT | signness);
		}/*}}}*/
	
	private:
		type_t* mDataType;
};/*}}}*/

class BinaryExpression;
class Call;
class Dummy;
class Global;
class NumericLiteral;
class Register;
class StackVariable;
class StringLiteral;
class TernaryExpression;
class UnaryExpression;

/**
 * Abstract base class for expression visitors
 */
class ExpressionVisitor
{
	public:
		virtual void Visit(BinaryExpression&)  = 0;
		virtual void Visit(Call&)              = 0;
		virtual void Visit(Dummy&)             = 0;
		virtual void Visit(Global&)            = 0;
		virtual void Visit(NumericLiteral&)    = 0;
		virtual void Visit(Register&)          = 0;
		virtual void Visit(StackVariable&)     = 0;
		virtual void Visit(StringLiteral&)     = 0;
		virtual void Visit(TernaryExpression&) = 0;
		virtual void Visit(UnaryExpression&)   = 0;
};


/**
 * Abstract base class for all expressions
 */
class Expression/*{{{*/
{
	public:
		enum ExpressionType
		{
			BINARY_EXPRESSION,
			CALL,
			DUMMY,
			GLOBAL,
			NUMERIC_LITERAL,
			REGISTER,
			STACK_VARIABLE,
			STRING_LITERAL,
			TERNARY_EXPRESSION,
			UNARY_EXPRESSION
		};

		ExpressionType Type() const throw() { return mType; }
		bool IsType(ExpressionType type) const throw() { return Type() == type; }

		TypeInformation& DataType()              { return mDataType; }

		virtual int SubExpressionCount()
		{
			return 0;
		}

		virtual Expression_ptr SubExpression(int index)
		{
			Expression_ptr e;
			return e;
		}

		virtual void SubExpression(int index, Expression_ptr e)
		{
		}

		virtual void GenerateCode(ostream& os, int level = 0)
		{
			os << "NYI";
		}

		virtual void Accept(ExpressionVisitor& visitor) = 0;

		/**
		 * Apply visitor to all sub-expressions with depth-first search
		 */
		virtual void AcceptDepthFirst(ExpressionVisitor& visitor)
		{
			for (int i = 0; i < SubExpressionCount(); i++)
				SubExpression(i)->AcceptDepthFirst(visitor);
			Accept(visitor);
		}

		static Expression_ptr FromOperand(insn_t& insn, int operand,
				TypeInformation* type = NULL);

		static bool Equal(Expression_ptr a, Expression_ptr b);

	protected:
		Expression(ExpressionType type)
			: mType(type)
		{
		}

	private:
		ExpressionType mType;
		TypeInformation mDataType;
};/*}}}*/

class UnaryExpression : public Expression/*{{{*/
{
	public:
		UnaryExpression(const char* operation, Expression_ptr operand)
			: Expression(UNARY_EXPRESSION), mOperation(operation),
				mOperand(operand)
		{}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual int SubExpressionCount()
		{
			return 1;
		}

		virtual Expression_ptr SubExpression(int /*index*/)
		{
			return mOperand;
		}

		virtual void SubExpression(int /*index*/, Expression_ptr e)
		{
			mOperand = e;
		}

		void Operand(Expression_ptr operand) { mOperand = operand; }
		Expression_ptr Operand() { return mOperand; }

		const char* Operation() { return mOperation.c_str(); }

#if 0
		virtual void DepthFirst(ExpressionVisitor& visitor)
		{
			mOperand->DepthFirst(visitor);
			Expression::DepthFirst(visitor);
		}
#endif
		
		virtual void GenerateCode(ostream& os, int level)
		{
			if (level > 0)
				os << '(';
			os << mOperation << ' ';
			mOperand->GenerateCode(os, level+1);
			if (level > 0)
				os << ')';
		}

	private:
		std::string mOperation;
		Expression_ptr mOperand;
};/*}}}*/

class BinaryExpression : public Expression/*{{{*/
{
	public:
		BinaryExpression(Expression_ptr first, const char* operation, 
				Expression_ptr second)
			: Expression(BINARY_EXPRESSION), mFirst(first), mOperation(operation),
				mSecond(second)
		{}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual int SubExpressionCount()
		{
			return 2;
		}

		virtual Expression_ptr SubExpression(int index)
		{
			if (0 == index)
				return mFirst;
			else
				return mSecond;
		}

		virtual void SubExpression(int index, Expression_ptr e)
		{
			if (0 == index)
				mFirst = e;
			else
				mSecond = e;
		}

		void First(Expression_ptr first) { mFirst = first; }
		Expression_ptr First() { return mFirst; }

		void Second(Expression_ptr second) { mSecond = second; }
		Expression_ptr Second() { return mSecond; }

		const char* Operation() { return mOperation.c_str(); }

#if 0
		virtual void DepthFirst(ExpressionVisitor& visitor)
		{
			mFirst->DepthFirst(visitor);
			mSecond->DepthFirst(visitor);
			Expression::DepthFirst(visitor);
		}
#endif
		
		virtual void GenerateCode(ostream& os, int level)
		{
			if (level > 0)
				os << '(';
			mFirst->GenerateCode(os, level+1);
			os << ' ' << mOperation << ' ';
			mSecond->GenerateCode(os, level+1);
			if (level > 0)
				os << ')';
		}

	private:
		Expression_ptr mFirst;
		std::string mOperation;
		Expression_ptr mSecond;
};/*}}}*/

/**
 * Used for the question-mark-colon operator
 */
class TernaryExpression : public Expression/*{{{*/
{
	public:
		TernaryExpression(Expression_ptr a, Expression_ptr b, Expression_ptr c)
			: Expression(TERNARY_EXPRESSION) 
		{
			mOperands[0] = a;
			mOperands[1] = b;
			mOperands[2] = c;
		}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual int SubExpressionCount()
		{
			return 3;
		}

		virtual Expression_ptr SubExpression(int index)
		{
			return mOperands[index];
		}

		virtual void SubExpression(int index, Expression_ptr e)
		{
			mOperands[index] = e;
		}
		
		virtual void GenerateCode(ostream& os, int level)
		{
			if (level > 0)
				os << '(';
			mOperands[0]->GenerateCode(os, level+1);
			os << " ? ";
			mOperands[1]->GenerateCode(os, level+1);
			os << " : ";
			mOperands[2]->GenerateCode(os, level+1);
			if (level > 0)
				os << ')';
		}

		static Expression_ptr Create(Expression_ptr a, Expression_ptr b, Expression_ptr c)
		{
			return Expression_ptr( new TernaryExpression(a, b, c) );
		}
		
	private:
		Expression_ptr mOperands[3];
};/*}}}*/

class Call : public Expression/*{{{*/
{
	private:
	
		struct GenerateCodeHelper
		{
			GenerateCodeHelper(std::ostream& os)
				: mOs(os), mFirst(true)
			{}
			
			void operator() (Expression_ptr e)
			{
				if (mFirst)
					mFirst = false;
				else
					mOs << ", ";

				e->GenerateCode(mOs, 0);
			}

			std::ostream& mOs;
			bool mFirst;
		};
	
	public:

		enum
		{
			UNKNOWN_PARAMETER_COUNT = -1,
			MAX_PARAMETERS = 15
		};
		
		Call(Expression_ptr function);
		virtual ~Call();

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual int SubExpressionCount()
		{
			return mSubExpressions.size();
		}

		virtual Expression_ptr SubExpression(int index)
		{
			return mSubExpressions[index];
		}

		virtual void SubExpression(int index, Expression_ptr e)
		{
			mSubExpressions[index] = e;
		}

		int ParameterCount() const { return mParameterCount; }

		void ParameterCountFromStack(int parameterCount)
		{
			if (ParameterCount() == UNKNOWN_PARAMETER_COUNT)
			{
				ParameterCount(parameterCount);
				CallingConvention(CM_CC_STDCALL);
			}
			else
			{
				msg("Warning: Ignoring parameter count from stack (previous=%i, suggestion=%i)\n",
						ParameterCount(), parameterCount);
			}
		}

		void ParameterCountFromCall(int parameterCount)
		{
			if (ParameterCount() == UNKNOWN_PARAMETER_COUNT)
			{
				ParameterCount(parameterCount);
				CallingConvention(CM_CC_CDECL);
			}
			else if (ParameterCount() == parameterCount)
			{
				// do nothing
			}
			else if (CM_CC_ELLIPSIS == mCallingConvention)
			{
				if (ParameterCount() > parameterCount)
					msg("Error: will not decrease number of parameters to ellipsis function (previous=%i, suggestion=%i)\n",
						ParameterCount(), parameterCount);
				else
					ParameterCount(parameterCount);
			}
			else
			{
				msg("Warning: Ignoring parameter count from call (previous=%i, suggestion=%i)\n",
						ParameterCount(), parameterCount);
			}
		}
	
#if 0
		void ParameterCount(int parameterCount) 
		{ 
			if (UNKNOWN_PARAMETER_COUNT != mParameterCount &&
					mParameterCount != parameterCount)
			{
				if (CM_CC_ELLIPSIS == mCallingConvention)
					msg("CM_CC_ELLIPSIS: "); /*&&
							parameterCount < mParameterCount)
				else*/
				{
					msg("Warning: changing call parameter count from %i to %i\n",
							mParameterCount, parameterCount);
				}
			}
			
			mParameterCount = parameterCount; 
			mSubExpressions.reserve(1+mParameterCount);
		}
#endif

		void AddParameter(Expression_ptr param)
		{
			if (mFinishedAddingParameters)
				msg("Warning! Adding parameters but were supposed to be finished\n");
			if (mSubExpressions.size() > mParameterCount)
				msg("Warning! Adding more parameters than parameter count\n");
			mSubExpressions.push_back(param);
		}

		bool IsFinishedAddingParameters()
		{
			return mFinishedAddingParameters;
		}

		void SetFinishedAddingParameters()
		{
			mFinishedAddingParameters = true;
		}

		void SetDataTypes();

		cm_t CallingConvention() const { return mCallingConvention; }
		void CallingConvention(cm_t cm) { mCallingConvention = cm & CM_CC_MASK; }

		bool IsCdecl() const 
		{
			switch (mCallingConvention)
			{
				case CM_CC_UNKNOWN:		/* possibly CDECL */
				case CM_CC_CDECL:
				case CM_CC_ELLIPSIS:
					return true;
	
				default:
					return false;
			}
		}

		/** special overloaded version */
		virtual void AcceptDepthFirst(ExpressionVisitor& visitor)
		{
			// First subexpression is function, but we want it to be examined last
			for (int i = 1; i < SubExpressionCount(); i++)
				SubExpression(i)->AcceptDepthFirst(visitor);
			Accept(visitor);
		}

		virtual void GenerateCode(ostream& os, int level)
		{
#if DUMP_DATA_TYPES
			os << '(';
			DataType().Write(os);
			os << ')';
#endif
			
			Function()->GenerateCode(os, level+1);
			os << '(';
			for_each(mSubExpressions.begin()+1, mSubExpressions.end(), 
					GenerateCodeHelper(os));
			os << ')';
		}

	protected:
		void ParameterCount(parameterCount)
		{
			mParameterCount = parameterCount; 
			mSubExpressions.reserve(1+mParameterCount);
		}

	private:
		Expression_ptr Function() { return mSubExpressions[0]; }
		ea_t mFunctionAddress;
		int mParameterCount;
		Expression_vector mSubExpressions;
		/* CM_CC_CDECL etc, from TYPEINF.HPP */
		cm_t mCallingConvention;
		bool mFinishedAddingParameters;

		type_t* mDataTypes     [MAX_PARAMETERS];
		char*   mParameterNames[MAX_PARAMETERS];
 
		void LoadTypeInformation();
		void LoadFunctionInformation();
};/*}}}*/

class NumericLiteral : public Expression/*{{{*/
{
	public:
		NumericLiteral(unsigned long value)
			: Expression(NUMERIC_LITERAL), mValue(value)
		{
			DataType().MakeInt();
		}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		unsigned long Value() { return mValue; }

		virtual void GenerateCode(ostream& os, int level)
		{
#if DUMP_DATA_TYPES
			os << '(';
			DataType().Write(os);
			os << ')';
#endif
			
			if (mValue >= 0xfffffff0)
				os << (signed long)mValue;
			else
			{
				if (mValue < 0x10)
					os << dec;
				else
					os << hex << "0x";
				os << mValue;
			}
		}

		static Expression_ptr Create(unsigned long value)
		{
			return Expression_ptr(new NumericLiteral(value));
		}

	private:
		unsigned long mValue;
};/*}}}*/

class StringLiteral : public Expression/*{{{*/
{
	public:
		StringLiteral(std::string value, ulong type)
			: Expression(STRING_LITERAL), mValue(value), mStringType(type)
		{
			DataType().MakeCharPointer();
		}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual void GenerateCode(ostream& os, int level)
		{
#if DUMP_DATA_TYPES
			os << '(';
			DataType().Write(os);
			os << ')';
#endif
				
			switch (mStringType)
			{
				case ASCSTR_UNICODE:
				case ASCSTR_ULEN2:
				case ASCSTR_ULEN4:
					os << "L";
					break;
			}
			os << '"' << EscapeAsciiString(mValue) << '"';
		}

		static Expression_ptr CreateFrom(ea_t address);
		static std::string GetString(ea_t address, ulong type);

		static std::string EscapeAsciiString(const std::string& str);
		static std::string EscapeAsciiString(const char* str);

	private:
		std::string mValue;
		ulong mStringType;
};/*}}}*/

class Register : public Expression/*{{{*/
{
	public:
		Register(unsigned short reg)
			: Expression(REGISTER), mRegister(reg)
		{
//			DataType().MakeInt();
		}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		unsigned short Index() { return mRegister; }

		unsigned short SimpleIndex()
		{
			// XXX: this is supposed to return AX for AL, etc
			return mRegister & 7;
		}

		virtual void GenerateCode(ostream& os, int level)
		{
#if DUMP_DATA_TYPES
			os << '(';
			DataType().Write(os);
			os << ')';
#endif
	
			os << Name(mRegister);
		}

		static const char* Name(int index);

		static Expression_ptr Create(unsigned short reg)
		{
			return Expression_ptr(new Register(reg));
		}

		static unsigned short Index(Expression_ptr e)
		{
			if (e->IsType(REGISTER))
				return static_cast<Register*>(e.get())->Index();
			else
				return (unsigned short)-1;
		}

	private:
		unsigned short mRegister;
		static const char* mName_x86_16bit[];
};/*}}}*/

/**
 * Abstract base class for Global and StackVariable
 */
class Location : public Expression/*{{{*/
{
	public:
		Location(ExpressionType type, std::string name, int index=0)
			: Expression(type), mName(name), mIndex(index)
		{}

		std::string Name() const throw() { return mName; }

		virtual void GenerateCode(ostream& os, int level)
		{
#if DUMP_DATA_TYPES
			os << '(';
			DataType().Write(os);
			os << ')';
#endif
	
			os << mName;
			if (mIndex)
			{
				os << '[' << mIndex << ']';
			}
		}

	private:
		int mIndex;
		std::string mName;
};/*}}}*/

class Global : public Location/*{{{*/
{
	public:
		Global(std::string name, int index=0, ea_t address=BADADDR)
			: Location(GLOBAL, name, index), mAddress(address)
		{
			DataType().MakeInt();
		}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		ea_t Address()
		{
			if (mAddress == BADADDR)
				mAddress = ::get_name_ea(BADADDR, Name().c_str());
			return mAddress;
		}


		static Expression_ptr CreateFrom(ea_t ea, ea_t from = BADADDR);
		static std::string GetName(ea_t ea, ea_t from = BADADDR);

	private:
		ea_t mAddress;
};/*}}}*/

class StackVariable : public Location/*{{{*/
{
	public:
		StackVariable(std::string name, int index=0)
			: Location(STACK_VARIABLE, name, index)
		{
			DataType().MakeInt();
		}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		static Expression_ptr CreateFrom(insn_t& insn, int operand);
		static std::string GetName(ea_t ea, int operand, int* pIndex=NULL);
};/*}}}*/

class Dummy : public Expression/*{{{*/
{
	public:
		Dummy()
			: Expression(DUMMY)
		{}

		virtual void Accept(ExpressionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual void GenerateCode(ostream& os, int level)
		{
			// empty
		}

		static Expression_ptr Create()
		{
			return Expression_ptr(new Dummy());
		}
};/*}}}*/

#endif

