// 
// Copyright (c) 2002 David Eriksson <david@2good.nu>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// $Id: instruction.hpp,v 1.1.1.1 2002/06/20 18:00:45 twogood Exp $
#ifndef _INSTRUCTION_HPP
#define _INSTRUCTION_HPP

#include <strstream>
#include "desquirr.hpp"

// Local includes

#include "expression.hpp"

class Expression;

typedef std::list<Expression*> ExpressionList;

typedef std::list<op_t> OperandList;

bool Equals(op_t& a, op_t& b);
void DumpInsn(insn_t& insn);
bool OperandIsRegister(insn_t& insn, int operand, int reg);
bool OperandIsImmediate(insn_t& insn, int operand, unsigned value);

/**
 * For Defined and Used 
 */
class BoolArray/*{{{*/
{
	protected:
		typedef unsigned long BITFIELD;

		BoolArray(BITFIELD bitfield)
			: mBitfield(bitfield)
		{}
		
	public:
		enum
		{
			SIZE = 16   // XXX: the size is just a nice number
		};
		
		BoolArray()
		{
			Clear();
		}

		BoolArray(BoolArray& other)
			: mBitfield(other.mBitfield)
		{
		}


		BoolArray operator ~ ()
		{
			return BoolArray(~mBitfield);
		}

		bool operator != (BoolArray& other)
		{
			return other.mBitfield != mBitfield;
		}

		void operator |= (BoolArray other)
		{
			mBitfield |= other.mBitfield;
		}

		BoolArray operator | (BoolArray other)
		{
			return BoolArray(mBitfield | other.mBitfield);
		}

		BoolArray operator & (BoolArray other)
		{
			return BoolArray(mBitfield & other.mBitfield);
		}

		bool Get(int i)
		{ 
			if (i >= 0 && i < SIZE) 
				return 0 != (mBitfield & POWER_OF_2[i]);
			else
				return false;
		}				

		void Set(int i)   
		{ 
			if (i >= 0 && i < SIZE)
				mBitfield |= POWER_OF_2[i];
		}

		void Clear(int i) 
		{ 
			mBitfield &= ~POWER_OF_2[i]; 
		}

		void Clear()
		{
			mBitfield = 0;
		}
			
		void Or(BoolArray& other)
		{
			mBitfield |= other.mBitfield;
		}

		int CountSet()
		{
			int count = 0;
			for (int i = 0; i < SIZE; i++)
				if (mBitfield & POWER_OF_2[i])
					count++;
			return count;
		}

		void Dump(std::ostream& os)
		{
			bool first = true;
			os << '{';
			for (int i = 0; i < SIZE; i++)
			{
				if (Get(i))
				{
					if (first)
						first = false;
					else
						os << ", ";

					os << Register::Name(i);
				}
			}
			os << '}';
		}

	private:
		BITFIELD mBitfield;

		static const int POWER_OF_2[SIZE];
};/*}}}*/

class Assignment;
class Case;
class ConditionalJump;
class Jump;
class Label;
class LowLevel;
class Push;
class Pop;
class Return;
class Switch;
class Throw;

/**
 * Abstract base class for Instruction visitors
 */
class InstructionVisitor
{
	public:
		virtual void Visit(Assignment&)      = 0;
		virtual void Visit(Case&)            = 0;
		virtual void Visit(ConditionalJump&) = 0;
		virtual void Visit(Jump&)            = 0;
		virtual void Visit(Label&)           = 0;
		virtual void Visit(LowLevel&)        = 0;
		virtual void Visit(Push&)            = 0;
		virtual void Visit(Pop&)             = 0;
		virtual void Visit(Return&)          = 0;
		virtual void Visit(Switch&)          = 0;
		virtual void Visit(Throw&)           = 0;

		// Helper functions when accepting node lists
		virtual void NodeBegin(Node_ptr) {}
		virtual void NodeEnd() {}
};

void Accept(Node_list& nodes, InstructionVisitor& visitor);
void Accept(Instruction_list& instructions, InstructionVisitor& visitor);


/**
 * an instruction
 */
class Instruction/*{{{*/
{
	public:
		enum InstructionType
		{
			ASSIGNMENT,
			CALL,
			CASE,
			CONDITIONAL_JUMP,
			JUMP,
			LABEL,
			LOW_LEVEL,
			PUSH,
			POP,
			RETURN,
			SWITCH,
			THROW,
			TO_BE_DELETED
		};

		enum OperandTypeValue
		{
			INVALID,
			DEFINITION,
			USE,
			USE_AND_DEFINITION
		};
		
		virtual ~Instruction()
		{}

#if 1
		virtual int OperandCount()
		{ 
			// Default implementation
			return 0; 
		}
		
		virtual Expression_ptr Operand(int index)
		{
			// Default implementation
			Expression_ptr result;
			return result;
		}

		virtual void Operand(int index, Expression_ptr e)
		{
			// Default implementation
		}

		virtual OperandTypeValue OperandType(int index)
		{
			return INVALID;
		}
#endif
	
		virtual ea_t Address() const { return mAddress; }
		virtual InstructionType Type() const { return mType; }
		
		virtual bool IsType(InstructionType type)
		{
			return Type() == type;
		}

		virtual void Accept(InstructionVisitor& visitor) = 0;

		/**
		 * Return true if the whole instruction can be removed
		 */
		virtual bool RemoveDefinition(unsigned short reg)
		{
			// Do nothing by default
			return false;
		}
		
		BoolArray& Definitions()       { return mDefinitions; }
		BoolArray& Uses()              { return mUses; }
		BoolArray& LastDefinitions()   { return mLastDefinitions; }
		BoolArray& FlagDefinitions()   { return mFlagDefinitions; }

		RegisterToAddress_map& DuChain() { return mDuChain; }

		void AddToDuChain(unsigned short reg, ea_t address)
		{
			mDuChain.insert(
					RegisterToAddress_pair(reg, address)
					);
		}

		bool DefinitionHasNoUses(unsigned short reg)
		{
			return mDuChain.count(reg) == 0;
		}

		void SetLastDefinition(unsigned short reg)
		{
			mLastDefinitions.Set(reg);
		}
		
		bool IsLastDefinition(unsigned short reg)
		{
			return mLastDefinitions.Get(reg);
		}

		bool MarkForDeletion()/*{{{*/
		{
			if (TO_BE_DELETED == mType)
			{
				return false;
			}
			else
			{			
				mType = TO_BE_DELETED;
				return true;
			}
		}/*}}}*/

		static insn_t GetLowLevelInstruction(ea_t address)
		{
			ua_ana0(address);
			return cmd;
		}
		
		static void FillList(func_t* function, Instruction_list& instructions);
		static void FindDefintionUseChains(Instruction_list& instructions);

	protected:
		Instruction(InstructionType type, ea_t ea)
			: mType(type), mAddress(ea)
		{}
		
	private:
		InstructionType mType;
		ea_t mAddress;
		BoolArray mDefinitions;
		BoolArray mUses;
		BoolArray mLastDefinitions;
		BoolArray mFlagDefinitions;
		RegisterToAddress_map mDuChain;
};/*}}}*/

/*
 * No-operand instructions
 */

class Label : public Instruction/*{{{*/
{
	public:
		Label(ea_t ea, const char* name)
			: Instruction(LABEL, ea), mName(name)
		{}

		const char* Name() { return mName.c_str(); }

		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

	private:
		std::string mName;
};/*}}}*/

class LowLevel : public Instruction /*{{{*/
{
	public:
		LowLevel(insn_t insn)
			: Instruction(LOW_LEVEL, insn.ea),
				mInsn(insn)
		{}

		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		insn_t& Insn() { return mInsn; }
		
		static bool Insn(Instruction_ptr instruction, insn_t& insn) 
		{ 
			if (instruction->IsType(Instruction::LOW_LEVEL))
			{
				insn = static_cast<LowLevel*>(instruction.get())->Insn();
				return true;
			}
		
			return false; 
		}
		
	private:
		insn_t mInsn;
		
};/*}}}*/


/**
 * Abstract base class for single-operand instructions
 */
class UnaryInstruction : public Instruction/*{{{*/
{
	public:
		void Operand(Expression_ptr operand) { mOperand = operand; }
		Expression_ptr Operand() { return mOperand; }

#if 1
		virtual int OperandCount()
		{ 
			return 1; 
		}
		
		virtual Expression_ptr Operand(int index)
		{
			Expression_ptr result;
			if (0 == index)
				result = mOperand;
			return result;
		}

		virtual void Operand(int index, Expression_ptr e)
		{
			if (0 == index)
				mOperand = e;
		}
#endif
	
	protected:
		UnaryInstruction(InstructionType type, ea_t ea, Expression_ptr operand)
			: Instruction(type, ea), mOperand(operand)
		{}

	private:
		Expression_ptr mOperand;
};/*}}}*/

/**
 * Abstract base class for double-operand instructions
 */
class BinaryInstruction : public Instruction/*{{{*/
{
	public:
		void First(Expression_ptr first) { mFirst = first; }
		Expression_ptr First() { return mFirst; }

		void Second(Expression_ptr second) { mSecond = second; }
		Expression_ptr Second() { return mSecond; }

#if 1
		virtual int OperandCount()
		{ 
			return 2; 
		}
		
		virtual Expression_ptr Operand(int index)
		{
			Expression_ptr result;
			if (0 == index)
				result = mFirst;
			else if (1 == index)
				result = mSecond;
			return result;
		}

		virtual void Operand(int index, Expression_ptr e)
		{
			if (0 == index)
				mFirst = e;
			else if (1 == index)
				mSecond = e;
		}
#endif
	
	protected:
		BinaryInstruction(InstructionType type, ea_t ea, 
				Expression_ptr first, Expression_ptr second)
			: Instruction(type, ea), mFirst(first), mSecond(second)
		{}

	private:
		Expression_ptr mFirst;
		Expression_ptr mSecond;
};/*}}}*/

/*
 * Single-operand instructions
 */

class Push : public UnaryInstruction/*{{{*/
{
	public:
		Push(ea_t ea, Expression_ptr operand)
			: UnaryInstruction(PUSH, ea, operand)
		{}

		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual OperandTypeValue OperandType(int index)
		{
			return USE;
		}
};/*}}}*/

class Pop : public UnaryInstruction/*{{{*/
{
	public:
		Pop(ea_t ea, Expression_ptr operand)
			: UnaryInstruction(POP, ea, operand)
		{}
		
		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}
		
		virtual OperandTypeValue OperandType(int index)
		{
			return DEFINITION;
		}
		
		virtual bool RemoveDefinition(unsigned short reg)
		{
			if (!Operand().get())
			{
				msg("%p Error: no operand!\n", Address());
				return false;
			}
			
			if (!Operand()->IsType(Expression::REGISTER))
			{
				msg("Error: trying to remove non-register defintion in POP\n");
				return false;
			}

			Register* expression = static_cast<Register*>(Operand().get());
			
			if (reg != expression->Index())
			{
				msg("Error: trying to remove a non-existing defintion in POP\n");
				return false;
			}

			Operand( Dummy::Create() );
			Definitions().Clear(reg);

			return false;
		}
};/*}}}*/

class Jump : public UnaryInstruction/*{{{*/
{
	public:
		Jump(ea_t ea, Expression_ptr destination)
			: UnaryInstruction(JUMP, ea, destination)
		{}
		
		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual OperandTypeValue OperandType(int index)
		{
			return USE;
		}
};/*}}}*/

class Return : public UnaryInstruction/*{{{*/
{
	public:
		Return(ea_t ea, Expression_ptr value)
			: UnaryInstruction(RETURN, ea, value)
		{}
		
		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual OperandTypeValue OperandType(int index)
		{
			return USE;
		}

};/*}}}*/

/*
 * Double-operand instructions
 */

class ConditionalJump : public BinaryInstruction/*{{{*/
{
	public:
		ConditionalJump(ea_t ea, 
				Expression_ptr condition, Expression_ptr destination)
			: BinaryInstruction(CONDITIONAL_JUMP, ea, condition, destination)
		{}

		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual OperandTypeValue OperandType(int index)
		{
			return USE;
		}

};/*}}}*/

class Assignment : public BinaryInstruction/*{{{*/
{
	public:
		Assignment(ea_t ea, Expression_ptr destination, Expression_ptr source)
			: BinaryInstruction(ASSIGNMENT, ea, destination, source)
		{}

		bool IsCall()
		{
			// TODO: make proper implementation
			return Second()->IsType(Expression::CALL);
		}

		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual OperandTypeValue OperandType(int index)
		{
			if (0 == index)
			{
				if (Operand(0)->IsType(Expression::UNARY_EXPRESSION))
				{
					// TODO: verify that the Operand is UnaryExpression("*", Register())
					// This is an indirect store, so we use the operand, not define it!
					return USE;
				}
				
				return DEFINITION;
			}
			else
				return USE;
		}

		virtual bool RemoveDefinition(unsigned short reg)
		{
			if (!First()->IsType(Expression::REGISTER))
			{
				msg("%p Error: trying to remove non-register defintion in assignment\n",
						Address());
				return false;
			}

			Register* expression = static_cast<Register*>(First().get());
			
			if (reg != expression->Index())
			{
				msg("Error: trying to remove a non-existing defintion\n");
				return false;
			}

			// Replace defintion with dummy instruction
			First( Dummy::Create() );
			Definitions().Clear(reg);

			// TODO: if the second operand does not contain a CALL expression we can 
			// return true here
			return !IsCall();
		}
};/*}}}*/


/*
 * Switch/case instructions
 */

class Switch : public UnaryInstruction/*{{{*/
{
	public:
		Switch(ea_t ea, Expression_ptr value, switch_info_t& si)
			: UnaryInstruction(SWITCH, ea, value), mSwitchInfo(si)
		{}

		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual OperandTypeValue OperandType(int index)
		{
			return USE;
		}

	private:
		switch_info_t mSwitchInfo;	
};/*}}}*/

class Case : public Instruction/*{{{*/
{
	public:
		Case(ea_t ea, unsigned int value)
			: Instruction(CASE, ea), mValue(value)
		{}

		unsigned int Value() { return mValue; }
		
		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

	private:
		unsigned int mValue;
};/*}}}*/

/**
 * A throw instruction can have either zero or one parameters
 */
class Throw : public Instruction/*{{{*/
{
	public:
		Throw(ea_t ea, Expression_ptr exception, std::string dataType)
			: Instruction(THROW, ea), mException(exception), mDataType(dataType)
		{}

		Throw(ea_t ea)
			: Instruction(THROW, ea)
		{}

		virtual void Accept(InstructionVisitor& visitor)
		{
			visitor.Visit(*this);
		}

		virtual int OperandCount()
		{ 
			return IsRethrow() ? 0 : 1; 
		}
		
		virtual Expression_ptr Operand(int index)
		{
			Expression_ptr result;
			if (0 == index)
				result = mException;
			return result;
		}

		virtual void Operand(int index, Expression_ptr e)
		{
			if (0 == index)
				mException = e;
		}

		bool IsRethrow()
		{
			return NULL == mException.get();
		}

		Expression_ptr Exception()
		{
			return mException;
		}
		
		std::string DataType() { return mDataType; }

	private:
		Expression_ptr mException;
		std::string mDataType;
};/*}}}*/


class ErasePool/*{{{*/
{
	private:
		typedef std::list<Instruction_list::iterator> IteratorList;

		IteratorList mIterators;
		Instruction_list& mInstructions;

		struct EraseHelper
		{
			EraseHelper(Instruction_list& instructions)
				: mInstructions(instructions)
			{}
		
			Instruction_list& mInstructions;
			
			void operator () (Instruction_list::iterator item)
			{
				mInstructions.erase(item);
			}
		};
			
	public:
		ErasePool(Instruction_list& instructions)
			: mInstructions(instructions)
		{}
			
		~ErasePool()
		{
			for_each(mIterators.begin(), mIterators.end(), EraseHelper(mInstructions));
		}

		void Erase(Instruction_list::iterator item)
		{
			if ((**item).MarkForDeletion())
			{
				mIterators.push_back(item);
			}
		}
};/*}}}*/


#endif // _INSTRUCTION_HPP

