// 
// Copyright (c) 2002 David Eriksson <david@2good.nu>
// 
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// $Id: node.hpp,v 1.1.1.1 2002/06/20 18:00:45 twogood Exp $
#ifndef _NODE_HPP
#define _NODE_HPP

//
// C++ includes
//
#include <list>
#include <algorithm>

//
// Boost includes
//
#include <boost/shared_ptr.hpp>

//
// Local includes
// 
#include "desquirr.hpp"
#include "instruction.hpp"

class Node/*{{{*/
{
	public:
		enum NodeType
		{
			CALL,								// XXX: not yet implemented
			CONDITIONAL_JUMP,
			FALL_THROUGH,
			JUMP,
			N_WAY,							// XXX: not yet implemented
			RETURN
		};

		Instruction_list& Instructions() { return mInstructions; }
		ea_t Address() { return mAddress; }

		BoolArray& Definitions() { return mDefinitions; }
		BoolArray& Uses()        { return mUses; }
		BoolArray& LiveIn()      { return mLiveIn; }
		BoolArray& LiveOut()     { return mLiveOut; }

		bool InLiveOut(short int reg)
		{
			return mLiveOut.Get(reg);
		}

		virtual int SuccessorCount() 
		{ 
			// default implementation
			return 0;
		}

		virtual ea_t SuccessorAddress(int index)
		{
			// default implementation
			return BADADDR;
		}

		virtual Node_ptr Successor(int index)
		{
			Node_ptr result;
			return result;
		}

		virtual bool ConnectSuccessor(int index, Node_ptr successor)
		{
			// default implementation
			return false;
		}

		static void CreateList(Instruction_list& instructions,
				Node_list& nodes);
		static void ConnectSuccessors(Node_list& nodes);

		static void FindDefintionUseChains(Node_list& nodes);
		static void LiveRegisterAnalysis(Node_list& nodes);

	protected:
		Node(NodeType type, 
				Instruction_list::iterator begin,
				Instruction_list::iterator end)
			: mAddress(BADADDR), mType(type)
		{
			for(Instruction_list::iterator item = begin;
					item != end; 
					item++)
			{
				mInstructions.push_back(*item);
			}

			if (mInstructions.size())
			{
				mAddress = (**mInstructions.begin()).Address();
			}
			else
			{
				msg("Warning! Empty node of type %i created!\n", mType);
			}
		}

	private:
		ea_t mAddress;
		NodeType mType;
		Instruction_list mInstructions;

		BoolArray mUses;
		BoolArray mDefinitions;
		BoolArray mLiveIn;
		BoolArray mLiveOut;
};/*}}}*/

class OneWayNode : public Node/*{{{*/
{
	protected:
		OneWayNode(NodeType type, ea_t successor,
				Instruction_list::iterator begin,
				Instruction_list::iterator end)
			: Node(type, begin, end), mSuccessorAddress(successor)
		{}

		virtual int SuccessorCount() 
		{ 
			return 1;
		}

		virtual ea_t SuccessorAddress(int index)
		{
			return 0 == index ? mSuccessorAddress : BADADDR;
		}

		virtual Node_ptr Successor(int index)
		{
			Node_ptr result;
			if (0 == index)
				result = mSuccessor;
			return result;
		}

		virtual bool ConnectSuccessor(int index, Node_ptr successor)
		{
			if (0 == index && successor->Address() == mSuccessorAddress)
			{
				mSuccessor = successor;
				return true;
			}
			else
				return false;
		}


	private:
		ea_t mSuccessorAddress;
		Node_ptr mSuccessor;
};/*}}}*/

class TwoWayNode : public Node/*{{{*/
{
	protected:
		TwoWayNode(NodeType type, ea_t successorA, ea_t successorB,
				Instruction_list::iterator begin,
				Instruction_list::iterator end)
			: Node(type, begin, end)
		{
			mSuccessorAddress[0] = successorA;
			mSuccessorAddress[1] = successorB;
		}

		virtual int SuccessorCount() 
		{ 
			return 2;
		}

		virtual ea_t SuccessorAddress(int index)
		{
			switch (index)
			{
				case 0:
				case 1:
					return mSuccessorAddress[index];
				default:
					return BADADDR;
			}
		}

		virtual Node_ptr Successor(int index)
		{
			Node_ptr result;
			switch (index)
			{
				case 0:
				case 1:
					result = mSuccessor[index];
			}
			return result;
		}

		virtual bool ConnectSuccessor(int index, Node_ptr successor)
		{
			switch (index)
			{
				case 0:
				case 1:
					if (successor->Address() == mSuccessorAddress[index])
					{
						mSuccessor[index] = successor;
						return true;
					}
					// fall through
					
				default:
					return false;
			}
		}


	private:
		ea_t mSuccessorAddress[2];
		Node_ptr mSuccessor[2];
};/*}}}*/


class JumpNode : public OneWayNode/*{{{*/
{
	public:
		JumpNode(ea_t destination,
				Instruction_list::iterator begin,
				Instruction_list::iterator end)
			: OneWayNode(JUMP, destination, begin, end)
		{}

		static Node_ptr CreateFrom(Instruction_ptr i,
				Instruction_list::iterator begin,
				Instruction_list::iterator end);
};/*}}}*/

class ConditionalJumpNode : public TwoWayNode /*{{{*/
{
	public:
		ConditionalJumpNode(
				ea_t destination, ea_t follower,
				Instruction_list::iterator begin,
				Instruction_list::iterator end)
			: TwoWayNode(CONDITIONAL_JUMP, destination, follower, begin, end)
		{}

		static Node_ptr CreateFrom(Instruction_ptr i,
				ea_t follower,
				Instruction_list::iterator begin,
				Instruction_list::iterator end);
};/*}}}*/

class FallThroughNode : public OneWayNode/*{{{*/
{
	public:
		FallThroughNode(
				ea_t follower,
				Instruction_list::iterator begin,
				Instruction_list::iterator end)
			: OneWayNode(FALL_THROUGH, follower, begin, end)
		{}
};/*}}}*/

class ReturnNode : public Node/*{{{*/
{
	public:
		ReturnNode(
				Instruction_list::iterator begin,
				Instruction_list::iterator end)
			: Node(RETURN, begin, end)
		{}
};/*}}}*/

#if 0
class CallNode : public OneWayNode
{
	public:
		CallNode(
				ea_t follower,
				Instruction_list::iterator begin,
				Instruction_list::iterator end)
			: CallNode(CALL, follower, begin, end)
		{}
};
#endif

#endif

